from stormvogel import pgc
from stormvogel.model import ModelType


def create_debugging_mdp():
    def available_actions(s: str):
        if s == ("init", "A"):
            return [["mec"], ["probs"]]
        return [[]]

    def delta(s: str, a: pgc.Action):
        if "mec" in a:
            return [(1, ("X", "mec"))]
        elif s == ("X", "mec"):
            return [(1, ("Y", "mec"))]
        elif s == ("Y", "mec"):
            return [(1, ("X", "mec"))]
        elif "probs" in a:
            return [(1, ("A", "E"))]
        elif s == ("A", "E"):
            return [(0.5, ("C", "D")), (0.5, ("B", "D"))]
        return [(1, s)]

    def labels(s):
        if isinstance(s, tuple):
            return list(s)
        return [s]

    return pgc.build_pgc(
        delta=delta,
        initial_state_pgc=("init", "A"),
        available_actions=available_actions,
        labels=labels,
        modeltype=ModelType.MDP,
    )



from stormvogel import pgc
from typing import Callable
import stormvogel.result
import stormvogel.model
import math
import imageio
import os

GRID_ACTION_LABEL_MAP = {0: "←", 1: "↓", 2: "→", 3: "↑", 4: "pickup", 5: "dropoff"}


def gymnasium_grid_to_stormvogel(
    env, action_label_map: dict[int, str] = GRID_ACTION_LABEL_MAP
):
    """Convert a FrozenLake, Taxi, or Cliffwalking gymnasium environment to an explicit stormvogel model."""
    TRANSITIONS = env.unwrapped.P
    NO_ACTIONS = env.action_space.n
    INV_MAP = {v: k for k, v in action_label_map.items()}
    ALL_ACTIONS = [[x] for x in action_label_map.values()]

    def action_numer_map(a: pgc.Action):
        return INV_MAP[a[0]]

    if "taxi" in env.spec.id.lower():
        # For Taxi, we need a special initial state that goes to every state. This is to account for the randomized starting position.
        init = pgc.State(n=-1, done=False)
    else:
        init = pgc.State(
            n=0, done=False
        )  # Otherwise, it's just 0 (Cliffwalking and FrozenLake).

    def available_actions(s: pgc.State):
        if s.n == -1:
            return [[]]
        return ALL_ACTIONS[:NO_ACTIONS]

    def delta(s: pgc.State, a: pgc.Action):
        if (
            s.n == -1
        ):  # Special taxi init state. It goes to every location that a passenger could spawn in. This should explore all states.
            # state = ((taxi_row * 5 + taxi_col) * 5 + passenger_location) * 4 + destination
            PLS = 4  # Number of passenger locations.
            return [(1 / PLS, pgc.State(n=x, done=False)) for x in range(PLS)]
        trans = TRANSITIONS[s.n][action_numer_map(a)]
        return list(map(lambda x: (x[0], pgc.State(n=int(x[1]), done=x[3])), trans))

    def rewards(s: pgc.State, a: pgc.Action) -> dict[str, stormvogel.model.Value]:
        if s.n == -1:
            return {"R": 0}
        reward = list(map(lambda x: x[2], TRANSITIONS[s.n][action_numer_map(a)]))[0]
        return {"R": reward}

    def labels(s: pgc.State):
        labels = [str(s.n), str(to_coordinate(s.n, env))]
        if s.n == get_target_state(env):
            labels.append("target")
        if s.done:
            labels.append("done")
        # labels.append("always")
        return labels

    return pgc.build_pgc(
        delta=delta,
        initial_state_pgc=init,
        available_actions=available_actions,
        labels=labels,
        rewards=rewards,
        modeltype=stormvogel.model.ModelType.MDP,
    )


def to_coordinate(s, env):
    """Calculate the state's coordinates. Works for FrozenLake, Cliffwalking, and Taxi"""
    num_states = env.observation_space.n
    grid_size = int(math.sqrt(num_states))
    x_target = int(s % grid_size)
    y_target = int(s // grid_size)
    return x_target, y_target


def to_state(x, y, env):
    """Calculate the state index from coordinates. Works for FrozenLake, CliffWalking, and Taxi."""
    num_states = env.observation_space.n
    grid_size = int(math.sqrt(num_states))
    return y * grid_size + x


def get_target_state(env):
    """Calculate the target state for an env. Works for FrozenLake and Cliffwalking"""
    return env.observation_space.n - 1


def to_gymnasium_scheduler(
    model: stormvogel.model.Model,
    scheduler: stormvogel.result.Scheduler
    | Callable[[stormvogel.model.State], stormvogel.model.Action],
    action_label_map: dict[int, str] = GRID_ACTION_LABEL_MAP,
) -> Callable[[int], int]:
    """Convert a stormvogel scheduler to a gymnasium scheduler (for a model that was converted using gymnasium_grid_to_stormvogel).
    Args:
        model: Stormvogel model.
        scheduler: Stormvogel scheduler.
        action_label_map: Map that you also used for the call in gymnasium_grid_to_stormvogel.
    Returns a gymnasium scheduler."""
    inv_map = {v: k for k, v in action_label_map.items()}

    def gymnasium_scheduler(env_sid: int):
        # TODO change this once pgc API features are a thing.
        model_state = model.get_states_with_label(str(int(env_sid)))[0]
        if isinstance(scheduler, stormvogel.result.Scheduler):
            choice = scheduler.get_choice_of_state(model_state)
        elif callable(scheduler):
            choice = scheduler(model_state)  # type: ignore
        return inv_map[list(choice.labels)[0]]

    return gymnasium_scheduler


def gymnasium_render_model_gif(
    env,
    gymnasium_scheduler: Callable[[int], int] | None = None,
    filename: str = "my_gif",
    max_length: int = 50,
    fps: int = 2,
    loop: int = 0,
) -> str:
    """Render a gymnasium model to a gif, using the gymnasium_scheduler (A map from state numbers to action numbers) to pick an action.
    Leave as None for a random action.

    Args:
        env: The gymnasium environment.
        gymnasium_scheduler: A function that takes a state number and returns an action number.
        filename: The name of the gif file to save.
        max_length: The maximum number of frames to render.
        fps: Frames per second for the gif.
        loop: Number of times to loop the gif (0 means loop forever).
    """
    frames = []  # List to store frames

    state, info = env.reset()
    for _ in range(max_length):
        frame = env.render()
        frames.append(frame)  # Store the frame

        action = (
            env.action_space.sample()
            if gymnasium_scheduler is None
            else gymnasium_scheduler(state)
        )
        state, reward, done, truncated, info = env.step(action)

        if done:
            frame = env.render()
            frames.append(frame)  # Render the last frame
            break  # Stop if the episode ends

    env.close()
    os.makedirs("gifs", exist_ok=True)
    # Save frames as a GIF
    imageio.mimsave(
        "gifs/" + filename + ".gif", frames, fps=fps, loop=loop
    )  # Adjust FPS as needed
    return "gifs/" + filename + ".gif"



from stormvogel.model import Model, ModelType, Branch, Transition


def create_lion_mdp():
    lion = Model(name="lion", model_type=ModelType.MDP)
    init = lion.get_initial_state()
    full = lion.new_state("full")
    satisfied = init
    hungry = lion.new_state("hungry :(")
    starving = lion.new_state("starving :((")
    dead = lion.new_state("dead")

    hunt = lion.new_action("hunt >:D")
    rawr = lion.new_action("rawr")

    init.set_transitions([(1, satisfied)])

    full.set_transitions(
        Transition(
            {
                hunt: Branch(
                    [
                        (0.5, satisfied),
                        (0.5, full),
                    ]
                ),
                rawr: Branch([(0.9, full), (0.1, satisfied)]),
            }
        )
    )

    satisfied.set_transitions(
        Transition(
            {
                hunt: Branch([(0.5, satisfied), (0.3, full), (0.2, hungry)]),
                rawr: Branch([(0.9, satisfied), (0.1, hungry)]),
            }
        )
    )

    hungry.set_transitions(
        Transition(
            {
                hunt: Branch(
                    [(0.2, full), (0.5, satisfied), (0.1, hungry), (0.2, starving)]
                ),
                rawr: Branch([(0.9, hungry), (0.1, starving)]),
            }
        )
    )

    starving.set_transitions(
        Transition(
            {
                hunt: Branch(
                    [(0.1, full), (0.5, satisfied), (0.2, hungry), (0.2, dead)]
                ),
                rawr: Branch([(0.9, starving), (0.1, dead)]),
            }
        )
    )
    lion.add_self_loops()

    reward_model = lion.add_rewards("R")
    reward_model.set_state_action_reward(full, rawr, 100)
    reward_model.set_unset_rewards(0)
    return lion



from bidict import bidict
from typing import Any

try:
    import stormpy
except ImportError:
    stormpy = None


def choice_mapping(sv_model, sp_model):
    """Return a bijective mapping between the stormvogel state-action pairs and the stormvogel model.
    WARNING: This function will be depricated later. It might also be faulty, I don't know :))"""
    res = bidict({})
    choice_id = 0
    for s in sv_model.states.values():
        for a in s.available_actions():
            res[s.id, a] = choice_id
            choice_id += 1
    return res


def to_bit_vector(state_set: set[int], model: Any):
    assert stormpy is not None
    return stormpy.BitVector(model.transition_matrix.nr_columns, list(state_set))



"""Contains the code responsible for saving/loading layouts and modifying them interactively."""

from typing import Any, Self

import numpy.typing as npt
import stormvogel.rdict

import os
import json
import copy

PACKAGE_ROOT_DIR = os.path.dirname(os.path.realpath(__file__))


class Layout:
    """Responsible for loading/saving layout jsons."""

    def __init__(
        self,
        path: str | None = None,
        path_relative: bool = True,
        layout_dict: dict | None = None,
    ) -> None:
        """Load a new Layout from a json file.
        Whenever keys are not present in the provided json file, their default values are used instead
        as specified in DEFAULTS (=layouts/default.json).

        Args:
            path (str, optional): Path to your custom layout file.
                Leave to None for an empty/default layout. Defaults to None.
            path_relative (bool, optional): If set to True, then stormvogel will look for a custom layout
                file relative to the current working directory. Defaults to True.
            layout_dict (dict, optional): If set, this dictionary is used as the layout instead of the
                file specified in path.
        """
        with open(os.path.join(PACKAGE_ROOT_DIR, "layouts/default.json")) as f:
            default_str = f.read()
        self.default_dict: dict = json.loads(default_str)

        if layout_dict is None:
            self.load(path, path_relative)
        else:
            self.layout: dict = stormvogel.rdict.merge_dict(
                self.default_dict, layout_dict
            )
            self.load_schema()

    def load_schema(self):
        """Load in the schema. Used for the layout editor. Stored as self.schema."""
        with open(os.path.join(PACKAGE_ROOT_DIR, "layouts/schema.json")) as f:
            schema_str = f.read()
        self.schema = json.loads(schema_str)

    def load(self, path: str | None = None, path_relative: bool = True):
        """Load the layout and schema file at the specified path.
        They are stored as self.layout and self.schema respectively."""
        if path is None:
            self.layout: dict = self.default_dict
        else:
            if path_relative:
                complete_path = os.path.join(os.getcwd(), path)
            else:
                complete_path = path
            with open(complete_path) as f:
                parsed_str = f.read()
            parsed_dict = json.loads(parsed_str)
            # Combine the parsed dict with default to fill missing keys as default values.
            self.layout: dict = stormvogel.rdict.merge_dict(
                self.default_dict, parsed_dict
            )
        self.load_schema()

    def add_active_group(self, group: str) -> None:
        """Make a group active if it is not already."""
        if group not in self.layout["edit_groups"]["groups"]:
            self.layout["edit_groups"]["groups"].append(group)

    def remove_active_group(self, group: str) -> None:
        """Make a group inactive if it is not already."""
        if group in self.layout["edit_groups"]["groups"]:
            self.layout["edit_groups"]["groups"].remove(group)

    def set_possible_groups(self, groups: set[str]):
        """Set the groups of states that the user can choose from under edit_groups."""
        self.schema["edit_groups"]["groups"]["__kwargs"]["allowed_tags"] = list(groups)

        # Save changes to the schema. The visualization object will handle putting nodes into the correct groups.
        groups2 = self.layout["edit_groups"]["groups"]
        self.schema[
            "groups"
        ] = {}  # empty the schema groups, to clear existing groups that we may not want
        for g in groups2:
            # For the settings themselves, we need to manually copy everything.
            layout_group_macro = copy.deepcopy(
                self.layout["__fake_macros"]["__group_macro"]
            )
            # Merge the macro with any existing changes.
            existing = self.layout["groups"][g] if g in self.layout["groups"] else {}
            self.layout["groups"][g] = stormvogel.rdict.merge_dict(
                layout_group_macro, existing
            )

            # For the schema, dict_editor already handles macros, so there is no need to do it manually here.
            if g not in self.schema["groups"]:
                self.schema["groups"][g] = {"__use_macro": "__group_macro"}

    def save(self, path: str, path_relative: bool = True) -> None:
        """Save this layout as a json file. Raises runtime error if a filename does not end in json, and OSError if file not found.

        Args:
            path (str): Path to your layout file.
            path_relative (bool, optional): If set to true, then stormvogel will create a custom layout
                file relative to the current working directory. Defaults to True.
        """
        if path[-5:] != ".json":
            raise RuntimeError("File name should end in .json")
        if path_relative:
            complete_path = os.path.join(os.getcwd(), path)
        else:
            complete_path = path
        with open(complete_path, "w") as f:
            json.dump(self.layout, f, indent=2)

    def set_value(self, path: list[str], value: Any):
        """Set a value in the layout. Also works if a key in the path does not exist."""
        stormvogel.rdict.rset(self.layout, path, value, create_new_keys=True)

    def __str__(self) -> str:
        return json.dumps(self.layout, indent=2)

    def copy_settings(self):
        """Copy some settings from one place in the layout to another place in the layout.
        They differ because visjs requires for them to be arranged a certain way which is not nice for an editor."""
        self.layout["physics"] = self.layout["misc"]["enable_physics"]

    def set_nx_pos(self, pos: dict[int, npt.NDArray], scale: float = 500) -> Self:
        """Apply NetworkX layout positions to this layout.

        Args:
            pos (dict[int, NDArray]): A dictionary of node positions from a NetworkX graph.
            scale (float): Scaling factor for the positions. Defaults to 500.
        """
        self.set_value(["physics"], False)
        self.set_value(
            ["positions"],
            {
                k: {"x": float(x * scale), "y": float(y * scale)}
                for k, (x, y) in pos.items()
            },
        )
        return self


# Define template layouts.
def DEFAULT():
    return Layout(
        os.path.join(PACKAGE_ROOT_DIR, "layouts/default.json"), path_relative=False
    )


def EXPLORE():
    default = DEFAULT()
    default.layout.layout["misc"]["explore"] = True
    return default


def SV():
    return Layout(
        os.path.join(PACKAGE_ROOT_DIR, "layouts/sv.json"), path_relative=False
    )



import stormvogel.model
from dataclasses import dataclass
from typing import cast, Any, Callable
import inspect


@dataclass
class State:
    """pgc state object. Can contain any number of any type of arguments"""

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"state({self.__dict__})"

    def __hash__(self):
        return hash(str(self.__dict__))

    def __eq__(self, other):
        if isinstance(other, State):
            return self.__dict__ == other.__dict__
        return False


type Action = list[str]


def valid_input(
    delta: Callable[[Any, Action], Any] | Callable[[Any], Any],
    initial_state_pgc,
    rewards: Callable[[Any, Action], dict[str, stormvogel.model.Value]]
    | Callable[[Any], dict[str, stormvogel.model.Value]]
    | None = None,
    labels: Callable[[Any], list[str] | str | None] | None = None,
    available_actions: Callable[[Any], list[Action]] | None = None,
    observations: Callable[[Any], int] | None = None,
    rates: Callable[[Any], float] | None = None,
    valuations: Callable[[Any], dict[str, float | int | bool]] | None = None,
    modeltype: stormvogel.model.ModelType = stormvogel.model.ModelType.MDP,
):
    """
    function that checks if the input for the pgc model builder is valid
    it will give a runtime error if it isn't.
    """

    supports_actions = modeltype in (
        stormvogel.model.ModelType.MDP,
        stormvogel.model.ModelType.POMDP,
        stormvogel.model.ModelType.MA,
    )

    # we first check if we have an available_actions function in case our model supports actions
    if supports_actions and available_actions is None:
        raise ValueError(
            "You have to provide an available actions function for models that support actions"
        )

    # and we check if we have an observations function in case our model is a POMDP
    if modeltype == stormvogel.model.ModelType.POMDP and observations is None:
        raise ValueError("You have to provide an observations function for pomdps")

    # we check if the provided functions have the right number of parameters
    if supports_actions:
        assert available_actions is not None
        sig = inspect.signature(available_actions)
        num_params = len(sig.parameters)
        if num_params != 1:
            raise ValueError(
                f"The available_actions function must take exactly one argument (state), but it takes {num_params} arguments"
            )

    sig = inspect.signature(delta)
    num_params = len(sig.parameters)
    if supports_actions:
        if num_params != 2:
            raise ValueError(
                f"Your delta function must take exactly two arguments (state and action), but it takes {num_params} arguments"
            )
    else:
        if num_params != 1:
            raise ValueError(
                f"Your delta function must take exactly one argument (state), but it takes {num_params} arguments"
            )

    if rewards is not None:
        sig = inspect.signature(rewards)
        num_params = len(sig.parameters)
        if supports_actions:
            if num_params != 2:
                raise ValueError(
                    f"The rewards function must take exactly two arguments (state, action), but it takes {num_params} arguments"
                )
        else:
            if num_params != 1:
                raise ValueError(
                    f"The rewards function must take exactly one argument (state), but it takes {num_params} arguments"
                )

    if labels is not None:
        sig = inspect.signature(labels)
        num_params = len(sig.parameters)
        if num_params != 1:
            raise ValueError(
                f"The labels function must take exactly one argument (state), but it takes {num_params} arguments"
            )

    if observations is not None:
        sig = inspect.signature(observations)
        num_params = len(sig.parameters)
        if num_params != 1:
            raise ValueError(
                f"The observations function must take exactly one argument (state), but it takes {num_params} arguments"
            )

    if valuations is not None:
        sig = inspect.signature(valuations)
        num_params = len(sig.parameters)
        if num_params != 1:
            raise ValueError(
                f"The valuations function must take exactly one argument (state), but it takes {num_params} arguments"
            )


def build_pgc(
    delta: Callable[[Any, Action], Any] | Callable[[Any], Any],
    initial_state_pgc: Any,
    rewards: Callable[[Any, Action], dict[str, stormvogel.model.Value]]
    | Callable[[Any], dict[str, stormvogel.model.Value]]
    | None = None,
    labels: Callable[[Any], list[str] | str | None] | None = None,
    available_actions: Callable[[Any], list[Action]] | None = None,
    observations: Callable[[Any], int] | None = None,
    rates: Callable[[Any], float] | None = None,
    valuations: Callable[[Any], dict[str, float | int | bool]] | None = None,
    modeltype: stormvogel.model.ModelType = stormvogel.model.ModelType.MDP,
    max_size: int = 10000,
) -> stormvogel.model.Model:
    """
    function that converts a delta function, an available_actions function an initial state and a model type
    to a stormvogel model

    this works analogous to a prism file, where the delta is the module in this case.

    (this function uses the pgc classes state and action instead of the ones from stormvogel.model)
    """

    def add_new_transitions(tuples, state):
        """
        helper function to add all the newly found transitions and states to the model
        """
        branch = []
        if tuples is not None:
            for tup in tuples:
                # in case only a state is provided, we assume the probability is 1
                if not isinstance(tup, tuple):
                    s = tup
                    val = 1
                else:
                    s = tup[1]
                    val = tup[0]

                if s not in state_lookup:
                    new_state = model.new_state(id=len(model.states))
                    state_lookup[s] = new_state
                    branch.append((val, new_state))
                    states_to_be_visited.append(s)
                else:
                    branch.append((val, state_lookup[s]))
        else:
            # if we have no return value, we add a self loop
            branch.append((1, state_lookup[state]))
        return branch

    valid_input(
        delta,
        initial_state_pgc,
        rewards,
        labels,
        available_actions,
        observations,
        rates,
        valuations,
        modeltype,
    )

    # we create the model with the given type and initial state
    model = stormvogel.model.new_model(modeltype=modeltype, create_initial_state=False)
    init = model.new_state(labels=["init"])

    # we continue calling delta and adding new states until no states are
    # left to be visited
    states_to_be_visited = [initial_state_pgc]
    state_lookup = {initial_state_pgc: init}
    while len(states_to_be_visited) > 0:
        state = states_to_be_visited.pop(0)
        transition = {}

        if model.supports_actions():
            # we loop over all available actions and call the delta function for each actions
            assert available_actions is not None
            actionslist = available_actions(state)

            if actionslist is None:
                raise ValueError(
                    f"On input {state}, the available actions function does not have a return value"
                )

            if not isinstance(actionslist, list):
                raise ValueError(
                    f"On input {state}, the available actions function does not return a list. Make sure to change it to [{actionslist}]"
                )

            for action in actionslist:
                try:
                    if action != []:
                        stormvogel_action = model.new_action(
                            frozenset(
                                action  # type: ignore
                            ),  # right now we only look at one label
                        )
                    else:
                        stormvogel_action = stormvogel.model.EmptyAction
                except RuntimeError:
                    if action != []:
                        stormvogel_action = model.get_action_with_labels(
                            frozenset(action)
                        )
                    else:
                        stormvogel_action = stormvogel.model.EmptyAction

                delta = cast(Callable[[Any, action], Any], delta)
                tuples = delta(state, action)

                if not isinstance(tuples, list) and tuples is not None:
                    raise ValueError(
                        f"On input pair {state} {action}, the delta function does not return a list. Make sure to change the format to [(<value>,<state>),...]"
                    )

                branch = add_new_transitions(tuples, state)

                if branch != []:
                    transition[stormvogel_action] = stormvogel.model.Branch(branch)
        else:
            delta = cast(Callable[[Any], Any], delta)
            tuples = delta(state)

            if not isinstance(tuples, list) and tuples is not None:
                raise ValueError(
                    f"On input {state}, the delta function does not return a list. Make sure to change the format to [(<value>,<state>),...]"
                )

            branch = add_new_transitions(tuples, state)

            if branch != []:
                transition[stormvogel.model.EmptyAction] = stormvogel.model.Branch(
                    branch
                )

        s = state_lookup[state]
        assert s is not None
        model.add_transitions(
            s,
            stormvogel.model.Transition(transition),
        )

        # if at some point we discovered more than max_size states, we complain
        if len(list(state_lookup)) > max_size:
            raise RuntimeError(
                f"The model you want te create has a very large amount of states (at least {max_size}), if you wish to proceed, set max_size to some larger number."
            )

    # we add the rewards
    if rewards is not None:
        if model.supports_actions():
            # we first create the right number of reward models
            assert available_actions is not None
            rewards = cast(
                Callable[[Any, action], dict[str, stormvogel.model.Value]], rewards
            )
            for reward in rewards(
                initial_state_pgc, available_actions(initial_state_pgc)[0]
            ).items():
                model.add_rewards(reward[0])

            # we take the initial state reward to compare later
            action = available_actions(initial_state_pgc)[0]
            initial_state_rewards = rewards(initial_state_pgc, action)

            for state, s in state_lookup.items():
                assert available_actions is not None
                for action in available_actions(state):
                    rewarddict = rewards(state, action)

                    # we check for the rewards when the function does not return a dict object
                    # or the length is not always the same
                    if rewarddict is None:
                        raise ValueError(
                            f"On input pair {state} {action}, the rewards function does not have a return value"
                        )

                    if not isinstance(rewarddict, dict):
                        raise ValueError(
                            f"On input pair {state} {action}, the rewards function does not return a dictionary. Make sure to change it to the format {{<rewardmodel>:<reward>,...}}"
                        )
                    if rewarddict.keys() != initial_state_rewards.keys():
                        raise ValueError(
                            "Make sure that the rewards function returns a dictionary with the same keys on each return"
                        )

                    assert s is not None
                    for index, reward in enumerate(rewarddict.items()):
                        a = model.get_action_with_labels(frozenset(action))
                        assert a is not None
                        model.rewards[index].set_state_action_reward(
                            s,
                            a,
                            reward[1],
                        )
        else:
            # we first create the right number of reward models
            rewards = cast(Callable[[Any], dict[str, stormvogel.model.Value]], rewards)
            for reward in rewards(initial_state_pgc).items():
                model.add_rewards(reward[0])

            initial_state_rewards = rewards(initial_state_pgc)
            for state, s in state_lookup.items():
                rewarddict = rewards(state)

                # we check for the rewards when the function does not return a dict object
                # or the length is not always the same
                if rewarddict is None:
                    raise ValueError(
                        f"On input {state}, the rewards function does not have a return value"
                    )

                if not isinstance(rewarddict, dict):
                    raise ValueError(
                        f"On input {state}, the rewards function does not return a dictionary. Make sure to change it to the format {{<rewardmodel name>:<reward>,...}}"
                    )
                if rewarddict.keys() != initial_state_rewards.keys():
                    raise ValueError(
                        "Make sure that the rewards function returns a dictionary with the same keys on each return"
                    )

                s = state_lookup[state]
                assert s is not None
                for index, reward in enumerate(rewarddict.items()):
                    model.rewards[index].set_state_reward(s, reward[1])

    # we add the observations
    if observations is not None:
        for state, s in state_lookup.items():
            # we check for the observations when it does not return an integer
            o = observations(state)
            if o is None:
                raise ValueError(
                    f"On input {state}, the observations function does not have a return value"
                )

            if not isinstance(o, int):
                raise ValueError(
                    f"On input {state}, the observations function does not return an integer"
                )

            s.set_observation(o)

    # we add the exit rates
    if rates is not None:
        for state, s in state_lookup.items():
            r = rates(state)
            if not isinstance(r, stormvogel.model.Value):
                raise ValueError(
                    f"On input {state}, the rates function does not return a number"
                )
            model.set_rate(s, r)

    # we add the valuations
    if valuations is not None:
        initial_state_valuations = valuations(initial_state_pgc)
        for state, s in state_lookup.items():
            valuation_list = valuations(state)
            if valuation_list is None:
                raise ValueError(
                    f"On input {state}, the valuations function does not have a return value"
                )

            if not isinstance(valuation_list, dict):
                raise ValueError(
                    f"On input {state}, the valuations function does not return a dictionary. Make sure to change the format to [<variable>: <value>,...]"
                )

            if valuation_list.keys() != initial_state_valuations.keys():
                raise RuntimeError(
                    "Make sure that you have a value for each variable in each state"
                )

            for val in valuation_list.values():
                if not (
                    isinstance(val, int)
                    or isinstance(val, bool)
                    or isinstance(val, float)
                ):
                    raise ValueError(
                        f"On input {state}, the dictionary that the valuations function returns contains a value {val} which is not of type int, float or"
                    )

            s.valuations = valuation_list

    # we add the labels
    if labels is not None:
        for state, s in state_lookup.items():
            labellist = labels(state)

            assert s is not None

            # if the labels function has no return value we assume this state simply has no labels
            if labellist is None:
                continue

            # we check for the labels when the function does not return a list object, or when
            # the list does not consist of strings
            if not isinstance(labellist, list):
                if not isinstance(labellist, str):
                    raise ValueError(
                        f"On input {state}, the labels function does not return a string or a list of strings"
                    )
                # if we don't get a list, we assume there is just one label
                if labellist not in s.labels:
                    s.add_label(labellist)
            else:
                for label in labellist:
                    if not isinstance(label, str):
                        raise ValueError(
                            f"On input {state}, the labels function does not return a string or a list of strings"
                        )
                    if label not in s.labels:
                        s.add_label(label)

    return model



from stormvogel import pgc
from stormvogel.model import ModelType


def create_end_components_mdp():
    init = pgc.State(x=["init"])

    def available_actions(s: pgc.State):
        if "init" in s.x:
            return [["one"], ["two"]]
        elif "mec1" in s.x or "mec2" in s.x:
            return [["one"], ["two"]]
        return [[]]

    def delta(s: pgc.State, a: pgc.Action):
        if "init" in s.x and "one" in a.labels:
            return [(0.5, pgc.State(x=["mec1"])), (0.5, pgc.State(x=["mec2"]))]
        elif "mec1" in s.x:
            return [(1, pgc.State(x=["mec2"]))]
        elif "mec2" in s.x:
            return [(1, pgc.State(x=["mec1"]))]
        elif "init" in s.x and "two" in a.labels:
            return [(1, pgc.State(x=["mec1"]))]
        return [(1, s)]

    def labels(s):
        return s.x

    mdp = pgc.build_pgc(
        delta=delta,
        initial_state_pgc=init,
        available_actions=available_actions,
        labels=labels,
        modeltype=ModelType.MDP,
    )
    return mdp
